@@ -4,6 +4,8 @@ use twitch_irc::{ClientConfig, SecureTCPTransport};
 
 #[tokio::main]
 pub async fn main() {
+    tracing_subscriber::fmt::init();
+
     // default configuration is to join chat as anonymous.
     let config = ClientConfig::default();
     let (mut incoming_messages, client) =
@@ -13,12 +15,15 @@ pub async fn main() {
     // otherwise they will back up.
     let join_handle = tokio::spawn(async move {
         while let Some(message) = incoming_messages.recv().await {
-            println!("Received message: {:?}", message);
+            tracing::info!("Received message: {:?}", message);
         }
     });
 
     // join a channel
-    client.join("sodapoppin".to_owned());
+    // This function only returns an error if the passed channel login name is malformed,
+    // so in this simple case where the channel name is hardcoded we can ignore the potential
+    // error with `unwrap`.
+    client.join("sodapoppin".to_owned()).unwrap();
 
     // keep the tokio executor alive.
     // If you return instead of waiting the background task will exit.
